// << ------------------------------------ >>
// << SLADE - SlayeR's LeetAss Doom Editor >>
// << By Simon Judd, 2004                  >>
// << ------------------------------------ >>
// << editor.cpp - 2d Editor functions     >>
// << ------------------------------------ >>

// Disable conversion warnings
#pragma warning (disable: 4244)

// INCLUDES ------------------------------ >>
#include "main.h"
#include "gui.h"
#include "menu.h"
#include "draw.h"
#include "map.h"
#include "checks.h"
#include "struct_3d.h"
#include "camera.h"
#include "tex_browse.h"
#include "console.h"
#include "misc.h"
#include "undoredo.h"

// VARIABLES ----------------------------- >>
int	zoom = 8;			// Zoom level (pixels per 32 map units)
int	xoff = 0;			// View x offset
int	yoff = 0;			// View y offset
int gridsize = 32;		// Grid size
int edit_mode = 1;		// Edit mode: 0=vertices, 1=lines, 2=sectors, 3=things
int	hilight_item = -1;	// The currently hilighted item (-1 for none)


int map_changed = 3;	// 0 = no change
						// 1 = changed, no node rebuild or struct rebuild needed
						// 2 = changed, no node rebuild needed
						// 3 = changed, node rebuild needed

bool browser = false;	// Is the texture browser active?
bool mix_tex = false;	// Mix textures & flats (zdoom)?

// Line drawing stuff
bool line_draw = false;		// Line drawing currently active?
pointlist_t ldraw_points;	// Points drawn (used to create verts & lines
numlist_t ldraw_lines;		// Lines drawn
numlist_t ldraw_verts;		// Vertices drawn

movelist_t	move_list;
numlist_t	select_order;

rect_t	sel_box(-1, -1, -1, -1);

CVAR(Bool, line_aa, false, CVAR_SAVE)
CVAR(Float, line_size, 1.0f, CVAR_SAVE)
CVAR(Bool, thing_sprites, false, CVAR_SAVE)

// Colours
rgba_t	col_hilight(255, 255, 0, 160);
rgba_t	col_selection(0, 140, 220, 255);
rgba_t	col_moving(255, 0, 0, 255);
rgba_t	col_tagged(0, 255, 0, 255);
rgba_t	col_cursor(100, 180, 255, 255);
rgba_t	col_vertex(0, 180, 0, 255);
rgba_t	col_line_solid(200, 200, 200, 255);
rgba_t	col_line_2s(110, 120, 130, 255);
rgba_t	col_line_monster(0, 120, 0, 255);
rgba_t	col_line_nofirst(255, 0, 0, 255);
rgba_t	col_linedraw(0, 255, 0, 255);
rgba_t	col_selbox(0, 140, 220, 100);
rgba_t	col_selbox_line(100, 180, 220, 200);
rgba_t	col_grid(0, 0, 255, 80);
rgba_t	col_64grid(200, 200, 200, 65);

GLuint	map_list;	// Display list for map
GLuint	grid_list;	// Display list for grid

// EXTERNAL VARIABLES -------------------- >>
EXTERN_CVAR(Int, vid_width)
EXTERN_CVAR(Int, vid_height)
extern Map map;
extern point2_t mouse;
extern bool menu, inbox, message, console;
extern Camera camera;
extern float scale;
extern thing_t last_thing;
extern stringlist_t popup_message;
extern string message_title;
extern TextureList tex_list;

// s_x: Returns the actual screen position of an x coordinate on the map
// ------------------------------------------------------------------ >>
int s_x(int x)
{
	return (vid_width / 2) + ((xoff * zoom) + (x * zoom) / MAJOR_UNIT);
}

// s_y: Returns the actual screen position of a y coordinate on the map
// ----------------------------------------------------------------- >>
int s_y(int y)
{
	return (vid_height / 2) + ((yoff * zoom) + (y * zoom) / MAJOR_UNIT);
}

point2_t s_p(point2_t point)
{
	return point2_t(s_x(point.x), s_y(point.y));
}

// m_x: Returns the map coordiate of an x position on the screen
// ---------------------------------------------------------- >>
int m_x(int x)
{
	return (int)((((float)x / (float)zoom) * 32.0f) - ((float)xoff * 32.0f) - ((((float)vid_width / 2.0f) / (float)zoom) * MAJOR_UNIT));
}

// m_y: Returns the map coordiate of a y position on the screen
// --------------------------------------------------------- >>
int m_y(int y)
{
	return (int)((((float)y / (float)zoom) * 32.0f) - ((float)yoff * 32.0f) - ((((float)vid_height / 2.0f) / (float)zoom) * MAJOR_UNIT));
}

point2_t m_p(point2_t point)
{
	return point2_t(m_x(point.x), m_y(point.y));
}

// difference: Finds the difference between 2 numbers
// ----------------------------------------------- >>
int difference(int v1, int v2)
{
	return max(v1, v2) - min(v1, v2);
}

// distance: Returns the distance between two points
// ---------------------------------------------- >>
double distance(int x1, int y1, int x2, int y2)
{
	double dist_x = (double)x2 - (double)x1;
	double dist_y = (double)y2 - (double)y1;

	return sqrt(dist_x * dist_x + dist_y * dist_y);
}

// midpoint: Returns the mid point of a line between two points
// --------------------------------------------------------- >>
point2_t midpoint(point2_t p1, point2_t p2)
{
	int mid_x = min(p1.x, p2.x) + (difference(p1.x, p2.x) / 2);
	int mid_y = min(p1.y, p2.y) + (difference(p1.y, p2.y) / 2);

	return point2_t(mid_x, mid_y);
}

// distance_to_line: Finds the shortest distance from a vertex to a line
// ------------------------------------------------------------------ >>
// (code borrowed & slightly modified from Doom Builder, by CodeImp)
double distance_to_line(int x1, int y1, int x2, int y2, int vx, int vy)
{
	double u;

	// Get length of linedef
	double ld = distance(x1, y1, x2, y2);

	// Check if not zero length
	if(ld)
	{
		// Get the intersection offset
		u = double((vx - x1) * (x2 - x1) + (vy - y1) * (y2 - y1)) / (ld * ld);

		// Limit intersection offset to the line
		double lbound = 1 / ld;
		double ubound = 1 - lbound;
		if(u < lbound) u = lbound;
		if(u > ubound) u = ubound;
	}
	else
	{
		u = 0;
	}

	// Calculate intersection point
	double ix = double(x1) + u * double(x2 - x1);
	double iy = double(y1) + u * double(y2 - y1);

	// Return distance between intersection and point
	// which is the shortest distance to the line
	return distance(double(vx), double(vy), ix, iy);
}

// in_rect: Determines if a point is within a rectangle
// ------------------------------------------------- >>
bool in_rect(int x1, int y1, int x2, int y2, int x, int y)
{
	return (x >= x1) && (x <= x2) && (y >= y1) && (y <= y2);
}

// snap_to_grid: Finds the nearest grid line to a point
// ------------------------------------------------- >>
int snap_to_grid(int pos)
{
	short upper, lower;

	for (int i = pos; i >= (pos - gridsize); i--)
	{
		if ((i % gridsize) == 0)
		{
			lower = i;
			break;
		}
	}

	for (i = pos; i < (pos + gridsize); i++)
	{
		if ((i % gridsize) == 0)
		{
			upper = i;
			break;
		}
	}

	if ((upper - pos) < (pos - lower))
		return upper;
	else
		return lower;

	return pos;
}

// snap_to_grid_custom: Finds the nearest specified grid line to a point
// ------------------------------------------------------------------ >>
int snap_to_grid_custom(int pos, int grid)
{
	short upper, lower;

	for (int i = pos; i >= (pos - grid); i--)
	{
		if ((i % grid) == 0)
		{
			lower = i;
			break;
		}
	}

	for (i = pos; i < (pos + grid); i++)
	{
		if ((i % grid) == 0)
		{
			upper = i;
			break;
		}
	}

	if ((upper - pos) < (pos - lower))
		return upper;
	else
		return lower;

	return pos;
}

// get_line_colour: Returns a colour depending on a line's flags
// ---------------------------------------------------------- >>
void get_line_colour(WORD l, rgba_t *colour)
{
	if (map.lines[l]->flags & LINE_IMPASSIBLE || map.lines[l]->side2 == -1)
		colour->set(col_line_solid);
	else
		colour->set(col_line_2s);

	if (map.lines[l]->side1 == -1)
		colour->set(col_line_nofirst);
}

// get_nearest_vertex: Gets the nearest vertex to a point
// --------------------------------------------------- >>
int get_nearest_vertex(int x, int y)
{
	double min_dist = 64;
	int vertex = -1;

	for (WORD v = 0; v < map.n_verts; v++)
	{
		double dist = distance(map.v_getpoint(v).x, map.v_getpoint(v).y, x, y);

		if (dist < min_dist)
		{
			min_dist = dist;
			vertex = v;
		}
	}

	return vertex;
}

// get_nearest_thing: Gets the nearest vertex to a point
// --------------------------------------------------- >>
int get_nearest_thing(int x, int y)
{
	double min_dist = 64;
	int thing = -1;

	for (WORD t = 0; t < map.n_things; t++)
	{
		double dist = distance(map.things[t]->x, map.things[t]->y, x, y);

		if (dist < min_dist)
		{
			min_dist = dist;
			thing = t;
		}
	}

	return thing;
}

// get_nearest_line: Gets the nearest line to a point
// (if any are closer than 64 units)
// ----------------------------------------------- >>
int get_nearest_line(int x, int y)
{
	double min_dist = 64;
	int line = -1;

	for (WORD l = 0; l < map.n_lines; l++)
	{
		rect_t r = map.l_getrect(l);

		double dist = distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), x, y);

		if (dist < min_dist)
		{
			min_dist = dist;
			line = l;
		}
	}

	return line;
}

// get_nearest_line_2: Gets the nearest line to a point
// ------------------------------------------------- >>
int get_nearest_line_2(int x, int y)
{
	double min_dist = -1;
	int line = -1;

	for (WORD l = 0; l < map.n_lines; l++)
	{
		rect_t r = map.l_getrect(l);

		double dist = distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), x, y);

		if (min_dist == -1 && map.l_getsector1(l) != -1)
		{
			min_dist = dist;
			line = l;
		}
		else if (dist < min_dist && map.l_getsector1(l) != -1)
		{
			min_dist = dist;
			line = l;
		}
	}

	return line;
}

// get_nearest_line_3: Gets the nearest line to a point (ignoring a sector)
// --------------------------------------------------------------------- >>
int get_nearest_line_3(int x, int y, int sector)
{
	double min_dist = -1;
	int line = -1;

	for (WORD l = 0; l < map.n_lines; l++)
	{
		rect_t r = map.l_getrect(l);

		double dist = distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), x, y);

		if (min_dist == -1 && map.l_getsector1(l) != -1)
		{
			if (map.l_getsector1(l) != sector && map.l_getsector2(l) != sector)
			{
				min_dist = dist;
				line = l;
			}
		}
		else if (dist < min_dist && map.l_getsector1(l) != -1)
		{
			if (map.l_getsector1(l) != sector && map.l_getsector2(l) != sector)
			{
				min_dist = dist;
				line = l;
			}
		}
	}

	return line;
}

// get_nearest_line_4: Gets the nearest line to a point (ignoring some lines)
// ----------------------------------------------------------------------- >>
int get_nearest_line_4(int x, int y, numlist_t *lines)
{
	double min_dist = -1;
	int line = -1;

	for (WORD l = 0; l < map.n_lines; l++)
	{
		if (!lines->exists(l))
		{
			rect_t r = map.l_getrect(l);

			double dist = distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), x, y);

			if (min_dist == -1 && map.l_getsector1(l) != -1)
			{
				min_dist = dist;
				line = l;
			}
			else if (dist < min_dist && map.l_getsector1(l) != -1)
			{
				min_dist = dist;
				line = l;
			}
		}
	}

	return line;
}

// clear_hilights: Clears hilighted map items depending on the editing mode
// --------------------------------------------------------------------- >>
void clear_hilights()
{
	hilight_item = -1;

	if (edit_mode == 0)
	{
		for (WORD v = 0; v < map.n_verts; v++)
			map.verts[v]->hilighted = false;
	}

	if (edit_mode == 1)
	{
		for (WORD l = 0; l < map.n_lines; l++)
			map.lines[l]->hilighted = false;
	}

	if (edit_mode == 2)
	{
		for (WORD s = 0; s < map.n_sectors; s++)
			map.sectors[s]->hilighted = false;
	}

	if (edit_mode == 3)
	{
		for (WORD t = 0; t < map.n_things; t++)
			map.things[t]->hilighted = false;
	}
}

// determine_line_side: Determines what side of a line a certain point is on
// ---------------------------------------------------------------------- >>
bool determine_line_side(int line, float x, float y)
{
	rect_t r;
	r.set(map.l_getrect(line));
	r.tl.y = -r.tl.y;
	r.br.y = -r.br.y;

	float side = (-y - (float)r.y1()) * float(r.x2() - r.x1()) - (x - (float)r.x1()) * float(r.y2() - r.y1());

	if (side > 0)
		return true;
	else
		return false;
}

// determine_line_side: Determines what side of a line a certain point is on
// ---------------------------------------------------------------------- >>
bool determine_line_side(rect_t r, float x, float y)
{
	r.tl.y = -r.tl.y;
	r.br.y = -r.br.y;

	float side = (-y - (float)r.y1()) * float(r.x2() - r.x1()) - (x - (float)r.x1()) * float(r.y2() - r.y1());

	if (side >= 0)
		return true;
	else
		return false;
}

// determine_line_side_f: Determines what side of a line a certain point is on
// (returns the actual value so we can test for being exactly on the line)
// ------------------------------------------------------------------------ >>
float determine_line_side_f(rect_t r, float x, float y)
{
	r.tl.y = -r.tl.y;
	r.br.y = -r.br.y;

	return (-y - (float)r.y1()) * float(r.x2() - r.x1()) - (x - (float)r.x1()) * float(r.y2() - r.y1());
}

// determine_line_side: Determines what side of a line a certain point is on
// ---------------------------------------------------------------------- >>
bool determine_line_side(float x1, float y1, float x2, float y2, float x, float y)
{
	float side = ((-y + y1) * (x2 - x1)) - ((x - x1) * (-y2 + y1));

	if (side > 0)
		return true;
	else
		return false;
}

// determine_sector: Determines what sector a point lies in
// (returns -1 if point is in the 'void')
int determine_sector(int x, int y)
{
	//draw_text(16, 16, COL_WHITE, 0, "%d %d", x, y);
	int line = get_nearest_line_2(x, y);

	if (line == -1)
		return -1;

	if (determine_line_side(line, x, y))
		return map.l_getsector1(line);
	else
		return map.l_getsector2(line);
}

/*
// determine_sector_ignore: Determines what sector a point lies in, ignoring the specified sector
// (returns -1 if point is in the 'void')
int determine_sector_ignore(int x, int y, int sector)
{
	//draw_text(16, 16, COL_WHITE, 0, "%d %d", x, y);
	int line = get_nearest_line_3(x, y, sector);

	if (line == -1)
		return -1;

	rect_t r;
	r.set(map.l_getrect(line));
	r.tl.y = -r.tl.y;
	r.br.y = -r.br.y;

	float side = float(-y - r.y1()) * float(r.x2() - r.x1()) - float(x - r.x1()) * float(r.y2() - r.y1());

	if (side >= 0)
		return map.l_getsector1(line);
	else
		return map.l_getsector2(line);
}
*/

int determine_sector_ignore(int x, int y, numlist_t *lines)
{
	//draw_text(16, 16, COL_WHITE, 0, "%d %d", x, y);
	int line = get_nearest_line_4(x, y, lines);

	if (line == -1)
		return -1;

	if (determine_line_side(line, x, y))
		return map.l_getsector1(line);
	else
		return map.l_getsector2(line);
}

// get_side_sector: Attempts to find what sector a line's side is in
// -------------------------------------------------------------- >>
int get_side_sector(int line, int side)
{
	rect_t linerect = map.l_getrect(line);
	point2_t mid = midpoint(linerect.tl, linerect.br);
	point2_t vec(linerect.x2() - linerect.x1(), linerect.y2() - linerect.y1());

	if (side == 2)
	{
		vec.x = -vec.x;
		vec.y = -vec.y;
	}

	int x = vec.y;
	int y = vec.x;
	x = -x;

	point2_t side_p(mid.x - x, mid.y - y);

	//printf("side %d\n", side);

	float min_dist = -1;
	int nearest_line = -1;
	point2_t nearest_midpoint;

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (l != line)
		{
			rect_t line_r = map.l_getrect(l);
			int x1 = line_r.x1();
			int x2 = line_r.x2();
			int y1 = line_r.y1();
			int y2 = line_r.y2();

			point2_t r1 = mid;
			point2_t r2 = side_p;

			float u_ray = (float(x2 - x1) * float(r1.y - y1) - float(y2 - y1) * float(r1.x - x1)) /
							(float(y2 - y1) * float(r2.x - r1.x) - float(x2 - x1) * float(r2.y - r1.y));
				
			float u_line = (float(r2.x - r1.x) * float(r1.y - y1) - float(r2.y - r1.y) * float(r1.x - x1)) /
							(float(y2 - y1) * float(r2.x - r1.x) - float(x2 - x1) * float(r2.y - r1.y));


			if ((u_ray >= 0) &&/* (u_ray < 1) && */(u_line >= 0) && (u_line <= 1))
			{
				if (u_ray < min_dist || min_dist == -1)
				{
					nearest_line = l;
					min_dist = u_ray;
					nearest_midpoint.set(midpoint(line_r.tl, line_r.br));
				}

				if (u_ray == min_dist)
				{
					double nearest_mid_dist = distance_to_line(linerect.x1(), linerect.y1(), linerect.x2(), linerect.y2(),
																nearest_midpoint.x, nearest_midpoint.y);

					point2_t this_midpoint(midpoint(line_r.tl, line_r.br));
					double this_mid_dist = distance_to_line(linerect.x1(), linerect.y1(), linerect.x2(), linerect.y2(),
															this_midpoint.x, this_midpoint.y);

					if (this_mid_dist < nearest_mid_dist)
					{
						nearest_line = l;
						min_dist = u_ray;
						nearest_midpoint.set(midpoint(line_r.tl, line_r.br));
					}
				}
			}
		}
	}

	if (nearest_line == -1)
	{
		//printf("No intersection\n");
		return -1;
	}
	else
	{
		if (determine_line_side(nearest_line, mid.x, mid.y) > 0.0f)
			return map.l_getsector1(nearest_line);
		else
			return map.l_getsector2(nearest_line);
	}
}

// get_hilight_item: Hilights the nearest item to the mouse pointer
// (if any items are close enough)
// ------------------------------------------------------------- >>
void get_hilight_item()
{
	clear_hilights();

	if (edit_mode == 0)
	{
		hilight_item = get_nearest_vertex(m_x(mouse.x), -m_y(mouse.y));

		if (hilight_item != -1)
			map.verts[hilight_item]->hilighted = true;
	}

	if (edit_mode == 1)
	{
		hilight_item = get_nearest_line(m_x(mouse.x), -m_y(mouse.y));

		if (hilight_item != -1)
			map.lines[hilight_item]->hilighted = true;
	}
	
	if (edit_mode == 2)
	{
		hilight_item = determine_sector(m_x(mouse.x), -m_y(mouse.y));

		if (hilight_item != -1)
			map.sectors[hilight_item]->hilighted = true;
	}

	if (edit_mode == 3)
	{
		hilight_item = get_nearest_thing(m_x(mouse.x), -m_y(mouse.y));

		if (hilight_item != -1)
			map.things[hilight_item]->hilighted = true;
	}
}

// selection: Returns true if any item is selected
// -------------------------------------------- >>
bool selection()
{
	if (edit_mode == 0)
	{
		for (WORD v = 0; v < map.n_verts; v++)
		{
			if (map.verts[v]->selected)
				return true;
		}
	}
	
	if (edit_mode == 1)
	{
		for (WORD l = 0; l < map.n_lines; l++)
		{
			if (map.lines[l]->selected)
				return true;
		}
	}

	if (edit_mode == 2)
	{
		for (WORD s = 0; s < map.n_sectors; s++)
		{
			if (map.s_isselected(s))
				return true;
		}
	}

	if (edit_mode == 3)
	{
		for (WORD t = 0; t < map.n_things; t++)
		{
			if (map.things[t]->selected)
				return true;
		}
	}

	return false;
}

// select_item: Selects/Deselects the currently highlighted item
// ---------------------------------------------------------- >>
void select_item()
{
	if (hilight_item == -1)
		return;

	// Vertices
	if (edit_mode == 0)
	{
		map.verts[hilight_item]->toggle_select();
		
		if (SELECTED(verts, hilight_item))
			select_order.add(hilight_item, false);
		else
			select_order.remove(hilight_item);
	}

	// Lines
	if (edit_mode == 1)
		map.lines[hilight_item]->toggle_select();

	// Sectors
	if (edit_mode == 2)
		map.sectors[hilight_item]->toggle_select();

	// Things
	if (edit_mode == 3)
		map.things[hilight_item]->toggle_select();

	update_map();
}

// select_items_box: Selects items within the selection box
// ----------------------------------------------------- >>
void select_items_box()
{
	int x1 = m_x(min(sel_box.x1(), sel_box.x2()));
	int x2 = m_x(max(sel_box.x1(), sel_box.x2()));
	int y1 = m_y(min(sel_box.y1(), sel_box.y2()));
	int y2 = m_y(max(sel_box.y1(), sel_box.y2()));

	// Vertices
	if (edit_mode == 0)
	{
		for (DWORD v = 0; v < map.n_verts; v++)
		{
			if (in_rect(x1, y1, x2, y2, map.verts[v]->x, -map.verts[v]->y))
			{
				map.verts[v]->select();
				select_order.add(v, true);
			}
		}
	}
	
	// Lines
	if (edit_mode == 1)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (in_rect(x1, y1, x2, y2, map.l_getrect(l).x1(), -map.l_getrect(l).y1())
				&& in_rect(x1, y1, x2, y2, map.l_getrect(l).x2(), -map.l_getrect(l).y2()))
				map.lines[l]->select();
		}
	}
	
	// Sectors
	if (edit_mode == 2)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			int sector1 = map.l_getsector1(l);
			int sector2 = map.l_getsector2(l);

			if (in_rect(x1, y1, x2, y2, map.l_getrect(l).x1(), -map.l_getrect(l).y1())
				&& in_rect(x1, y1, x2, y2, map.l_getrect(l).x2(), -map.l_getrect(l).y2()))
			{
				if (sector1 != -1)
					map.sectors[sector1]->select();

				if (sector2 != -1)
					map.sectors[sector2]->select();
			}
		}
	}
	
	// Things
	if (edit_mode == 3)
	{
		for (DWORD t = 0; t < map.n_things; t++)
		{
			if (in_rect(x1, y1, x2, y2, map.things[t]->x, -map.things[t]->y))
				map.things[t]->select();
		}
	}

	update_map();
}

// clear_selection: Clears selected map items depending on the editing mode
// --------------------------------------------------------------------- >>
void clear_selection()
{
	hilight_item = -1;

	if (edit_mode == 0)
	{
		for (WORD v = 0; v < map.n_verts; v++)
			map.verts[v]->deselect();
	}

	if (edit_mode == 1)
	{
		for (WORD l = 0; l < map.n_lines; l++)
			map.lines[l]->deselect();
	}

	if (edit_mode == 2)
	{
		for (WORD s = 0; s < map.n_sectors; s++)
			map.sectors[s]->deselect();
	}

	if (edit_mode == 3)
	{
		for (WORD t = 0; t < map.n_things; t++)
			map.things[t]->deselect();
	}

	select_order.clear();
	update_map();
}

// draw_hilight: Draws the currently hilighted item
// --------------------------------------------- >>
void draw_hilight()
{
	if (edit_mode == 0)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		draw_point(s_x(map.verts[hilight_item]->x), s_y(-map.verts[hilight_item]->y), 10, col_hilight);
	}

	if (edit_mode == 1)
	{
		rect_t rect = map.l_getsrect(hilight_item);
		rect_t rect2;

		int xm = rect.x1() + ((rect.x2() - rect.x1()) / 2);
		int ym = rect.y1() + ((rect.y2() - rect.y1()) / 2);
		int xm2, ym2;

		xm2 = xm - ((rect.y2() - rect.y1()) / 8);
		ym2 = ym + ((rect.x2() - rect.x1()) / 8);

		rect2.set(xm, ym, xm2, ym2);

		glLineWidth(3.0f);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		draw_line(rect, col_hilight, true);
		draw_line(rect2, col_hilight, true);
	}

	if (edit_mode == 2)
	{
		for(DWORD l = 0; l < map.n_lines; l++)
		{
			short sector1 = map.l_getsector1(l);
			short sector2 = map.l_getsector2(l);

			rect_t rect = map.l_getsrect(l);

			if (map.s_ishilighted(sector1)
				|| map.s_ishilighted(sector2))
			{
				glLineWidth(3.0f);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
				draw_line(rect, col_hilight, true);
			}
		}
	}

	if (edit_mode == 3)
	{
		int r = (map.things[hilight_item]->radius * zoom) / MAJOR_UNIT;
		point2_t p;
		p.set(s_x(map.things[hilight_item]->x), s_y(-map.things[hilight_item]->y));

		if (map.things[hilight_item]->radius == -1)
			r = 8 * zoom / MAJOR_UNIT;

		draw_rect(rect_t(p.x - r, p.y - r, p.x + r, p.y + r), col_hilight, true);
	}
}

// draw_basic_lines: Draws lines without selections/hilights or side markers
// ---------------------------------------------------------------------- >>
void draw_basic_lines()
{
	rgba_t colour(0, 0, 0, 255);

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		rect_t rect = map.l_getsrect(l);

		if (!((rect.x1() < 0 && rect.x2() < 0) ||
			(rect.x1() > vid_width && rect.x2() > vid_width) ||
			(rect.y1() < 0 && rect.y2() < 0) ||
			(rect.y1() > vid_height && rect.y2() > vid_height)))
		{
			glLineWidth(line_size);
			get_line_colour(l, &colour);
			draw_line(rect, colour, line_aa);
		}
	}
}

// draw_lines: Draws lines with selections, hilights and side_markers
// --------------------------------------------------------------- >>
void draw_lines()
{
	rgba_t colour;

	draw_basic_lines();

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		glLineWidth(line_size);

		rect_t rect = map.l_getsrect(l);
		rect_t rect2;

		if (!((rect.x1() < 0 && rect.x2() < 0) ||
			(rect.x1() > vid_width && rect.x2() > vid_width) ||
			(rect.y1() < 0 && rect.y2() < 0) ||
			(rect.y1() > vid_height && rect.y2() > vid_height)))
		{
			point2_t mid = midpoint(rect.tl, rect.br);
			point2_t vec(rect.x2() - rect.x1(), rect.y2() - rect.y1());
			int x = -vec.y / 8;
			int y = vec.x / 8;
			rect2.set(mid.x, mid.y, mid.x + x, mid.y + y);

			get_line_colour(l, &colour);
			draw_line(rect2, colour, line_aa);

			if (map.lines[l]->moving)
			{
				glLineWidth(5.0f);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
				draw_line(rect, col_moving, true);
				draw_line(rect2, col_moving, true);
			}
			else
			{
				if (map.lines[l]->selected)
				{
					glLineWidth(5.0f);
					glBlendFunc(GL_SRC_ALPHA, GL_ONE);
					draw_line(rect, col_selection, true);
					draw_line(rect2, col_selection, true);
				}
			}
		}

		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
}

// draw_vertices: Draws map vertices
// ------------------------------ >>
void draw_vertices()
{
	int size = 6;

	glEnable(GL_POINT_SMOOTH);

	for(DWORD v = 0; v < map.n_verts; v++)
	{
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		point2_t vert;
		vert.set(map.v_getspoint(v));

		// Deafults
		size = 6;

		if (vert.x >= 0 && vert.x <= vid_width && vert.y >= 0 && vert.y <= vid_height)
		{
			draw_point(vert.x, vert.y, size, col_vertex);

			if (map.verts[v]->moving)
			{
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
				draw_point(vert.x, vert.y, 10, col_moving);
			}
			else
			{
				if (map.verts[v]->selected)
				{
					glBlendFunc(GL_SRC_ALPHA, GL_ONE);
					draw_point(vert.x, vert.y, 10, col_selection);
				}
			}
		}
	}

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

// draw_sectors: Draws the map sectors (hilight/selections)
// ----------------------------------------------------- >>
void draw_sectors()
{
	rgba_t colour;

	for(DWORD l = 0; l < map.n_lines; l++)
	{
		short sector1 = map.l_getsector1(l);
		short sector2 = map.l_getsector2(l);

		rect_t rect = map.l_getsrect(l);

		if (map.s_ismoving(sector1)
			|| map.s_ismoving(sector2))
		{
			glLineWidth(5.0f);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE);
			draw_line(rect, col_moving, true);
		}
		else
		{
			if (map.s_isselected(sector1)
				|| map.s_isselected(sector2))
			{
				glLineWidth(5.0f);
				glBlendFunc(GL_SRC_ALPHA, GL_ONE);
				draw_line(rect, col_selection, true);
			}
		}
	}

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

// draw_things: Draws map things
// -------------------------- >>
void draw_things()
{
	for (DWORD t = 0; t < map.n_things; t++)
	{
		point2_t p;
		p.set(s_x(map.things[t]->x), s_y(-map.things[t]->y));

		if (p.x > 0 && p.x < vid_width && p.y > 0 && p.y < vid_height)
		{
			glEnable(GL_POINT_SMOOTH);
			int r = (map.things[t]->radius * zoom) / MAJOR_UNIT;

			if (map.things[t]->radius == -1)
				r = 8 * zoom / MAJOR_UNIT;

			if (thing_sprites)
			{
				int width = map.things[t]->tdef->sprite->width * zoom / MAJOR_UNIT;
				int height = map.things[t]->tdef->sprite->height * zoom / MAJOR_UNIT;

				if (map.things[t]->tdef->sprite->name[0] == '_')
					width = height = 16;

				draw_texture(map.things[t]->tdef->sprite, p.x - width / 2, p.y - height / 2, width, height, COL_WHITE);
			}
			else
			{
				if (map.things[t]->tdef->sprite->name[0] == '_')
					draw_texture(map.things[t]->tdef->sprite->name, p.x - r, p.y - r, r + r, r + r, map.things[t]->colour);
				else
					draw_texture("_thing", p.x - r, p.y - r, r + r, r + r, map.things[t]->colour);
			}

			if (map.things[t]->draw_angle)
			{
				int x2, y2;

				// east
				if (map.things[t]->angle == 0)			{ x2 = p.x + r; y2 = p.y; }
				// northeast
				else if (map.things[t]->angle == 45)	{ x2 = p.x + r; y2 = p.y - r; }
				// north
				else if (map.things[t]->angle == 90)	{ x2 = p.x; y2 = p.y - r; }
				// northwest
				else if (map.things[t]->angle == 135)	{ x2 = p.x - r; y2 = p.y - r; }
				// west
				else if (map.things[t]->angle == 180)	{ x2 = p.x - r; y2 = p.y; }
				// southwest
				else if (map.things[t]->angle == 225)	{ x2 = p.x - r; y2 = p.y + r; }
				// south
				else if (map.things[t]->angle == 270)	{ x2 = p.x; y2 = p.y + r; }
				// southeast
				else if (map.things[t]->angle == 315)	{ x2 = p.x + r; y2 = p.y + r; }
				// Invalid angle
				else	{ x2 = p.x; y2 = p.y; }

				glLineWidth(2.0f);
				if (thing_sprites)
				{
					draw_point(p.x, p.y, 8 * zoom / MAJOR_UNIT, COL_WHITE);
					draw_line(rect_t(p.x, p.y, x2, y2), rgba_t(255, 255, 255, 200), line_aa);
				}
				else
					draw_line(rect_t(p.x, p.y, x2, y2), rgba_t(0, 0, 0, 200), line_aa);
				glLineWidth(1.0f);
			}

			glBlendFunc(GL_SRC_ALPHA, GL_ONE);

			// Draw moving indicator (if moving)
			if (map.things[t]->moving)
				draw_rect(rect_t(p.x - r, p.y - r, p.x + r, p.y + r), col_moving, true);
			else
			{
				// Draw selection (if any)
				if (map.things[t]->selected)
					draw_rect(rect_t(p.x - r, p.y - r, p.x + r, p.y + r), col_selection, true);
			}

			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
		}
	}
}

// update_grid: Updates the grid display list
// --------------------------------------- >>
void update_grid()
{
	int grid_hidelevel;

	glDeleteLists(grid_list, 1);
	grid_list = glGenLists(1);
	glNewList(grid_list, GL_COMPILE);

	glDisable(GL_LINE_SMOOTH); // AA is unnecessary (grid lines are not diagonal)
	glLineWidth(1.0f);

	grid_hidelevel = 64;

	int start_x = m_x(0);
	int end_x = m_x(vid_width);
	int start_y = m_y(0);
	int end_y = m_y(vid_height);

	// If the grid should be drawn (not to small)
	if (gridsize >= (grid_hidelevel / zoom) && gridsize > 1)
	{
		for (int x = start_x; x < end_x; x++)
		{
			if (x % gridsize == 0)
				draw_line(rect_t(s_x(x), 0, s_x(x), vid_height), col_grid, false);
		}

		for (int y = start_y; y < end_y; y++)
		{
			if (y % gridsize == 0)
				draw_line(rect_t(0, s_y(y), vid_width, s_y(y)), col_grid, false);
		}
	}

	// Don't draw 64x64 grid if it'll be too small
	if (64 < (grid_hidelevel / zoom))
	{
		glEndList();
		return;
	}

	// Don't draw 64x64 grid if the actual grid size is >=
	if (gridsize >= 64)
	{
		glEndList();
		return;
	}

	//glLineStipple(4, 0xAAAA);
	//glEnable(GL_LINE_STIPPLE);

	for (int x = start_x; x < end_x; x++)
	{
		if (x % 64 == 0)
			draw_line(rect_t(s_x(x), 0, s_x(x), vid_height), col_64grid, false);
	}

	for (int y = start_y; y < end_y; y++)
	{
		if (y % 64 == 0)
			draw_line(rect_t(0, s_y(y), vid_width, s_y(y)), col_64grid, false);
	}

	//glDisable(GL_LINE_STIPPLE);
	glEndList();
}

// draw_info: Draws gui info boxes depending on edit mode
// --------------------------------------------------- >>
void draw_info()
{
	if (edit_mode == 0)
		gui_vertexinfo(hilight_item);

	if (edit_mode == 1)
		gui_lineinfo(hilight_item);

	if (edit_mode == 2)
		gui_sectorinfo(hilight_item);

	if (edit_mode == 3)
		gui_thinginfo(hilight_item);
}

// draw_drawlines: Draws dashed lines for line drawing
// ------------------------------------------------ >>
void draw_drawlines()
{
	glEnable(GL_POINT_SMOOTH);

	if (ldraw_points.n_points == 0)
	{
		draw_point(s_x(snap_to_grid(m_x(mouse.x))), s_y(snap_to_grid(m_y(mouse.y))), 8, col_linedraw);
		return;
	}
	else
	{
		glLineWidth(line_size);
		glLineStipple(4, 0xBBBB);
		glEnable(GL_LINE_STIPPLE);
		point2_t p1, p2;

		for (int p = 1; p < ldraw_points.n_points; p++)
		{
			p1.set(s_x(ldraw_points.points[p - 1].x), s_y(-ldraw_points.points[p - 1].y));
			p2.set(s_x(ldraw_points.points[p].x), s_y(-ldraw_points.points[p].y));

			draw_line(rect_t(p1, p2), col_linedraw, true);
			draw_point(p1.x, p1.y, 6, col_linedraw);
			draw_text(midpoint(p1, p2).x, midpoint(p1, p2).y - 4, col_linedraw, 1, "%d", (int)distance(m_x(p1.x), m_y(p1.y), m_x(p2.x), m_y(p2.y)));
		}
		
		if (ldraw_points.n_points == 1)
			p2.set(s_x(ldraw_points.points[0].x), s_y(-ldraw_points.points[0].y));

		p1.set(snap_to_grid(m_x(mouse.x)), snap_to_grid(m_y(mouse.y)));
		draw_line(rect_t(s_p(p1), p2), col_linedraw, true);
		draw_point(s_x(p1.x), s_y(p1.y), 8, col_linedraw);
		draw_point(p2.x, p2.y, 6, col_linedraw);

		draw_text(midpoint(s_p(p1), p2).x, midpoint(s_p(p1), p2).y, col_linedraw, 1, "%d", (int)distance(p1.x, p1.y, m_x(p2.x), m_y(p2.y)));

		glDisable(GL_LINE_STIPPLE);
	}
}

// update_map: Updates the map display list
// ------------------------------------- >>
void update_map()
{
	glDeleteLists(map_list, 1);
	map_list = glGenLists(1);
	glNewList(map_list, GL_COMPILE);

	// Draw map objects
	if (edit_mode == 0)	// Vertices
	{
		draw_basic_lines();
		draw_vertices();
	}

	if (edit_mode == 1)	// Lines
		draw_lines();

	if (edit_mode == 2) // Sectors
	{
		draw_basic_lines();
		draw_sectors();
	}

	if (edit_mode == 3) // Things
	{
		draw_basic_lines();
		draw_things();
	}

	glEndList();
}

// draw_2d_screen: Draws the 2d editor screen
// -------------------------------------- >>
void draw_2d_screen()
{
	// Clear screen
	glLoadIdentity();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (browser)
	{
		draw_browser();
		SDL_GL_SwapBuffers();
		return;
	}

	// Draw the grid
	glCallList(grid_list);

	// Draw the map/objects
	glCallList(map_list);

	// Draw hilight
	if (hilight_item != -1 && !line_draw && !menu)
		draw_hilight();

	// Draw selection box
	glLineWidth(2.0f);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	draw_rect(sel_box, col_selbox, true);
	draw_rect(sel_box, col_selbox_line, false);

	// Draw gui
	gui_menu();
	gui_status_bar();
	draw_info();

	if (menu)
		menu_draw();

	if (message)
		gui_popupmessage(popup_message, message_title.c_str());

	if (line_draw)
		draw_drawlines();

	// Draw mouse pointer
	if (!inbox && !message)
		draw_texture("_cursor", mouse.x, mouse.y, (vid_width * 0.024), (vid_height * 0.032), col_cursor);

	SDL_GL_SwapBuffers();
}

// init_map: Focuses on the center of the map & inits 3d camera
// --------------------------------------------------------- >>
void init_map()
{
	int max_x = 0;
	int min_x = 0;
	int max_y = 0;
	int min_y = 0;
	
	for (DWORD v = 0; v < map.n_verts; v++)
	{
		if (map.verts[v]->x < min_x)
			min_x = map.verts[v]->x;
		
		if (map.verts[v]->x > max_x)
			max_x = map.verts[v]->x;
		
		if (map.verts[v]->y < min_y)
			min_y = map.verts[v]->y;
		
		if (map.verts[v]->y > max_y)
			max_y = map.verts[v]->y;
	}
	
	xoff = -(min_x + ((max_x - min_x) / 2)) / MAJOR_UNIT;
	yoff = (min_y + ((max_y - min_y) / 2)) / MAJOR_UNIT;

	// Init camera
	point3_t pos;
	point3_t view;
	pos.z = 0.0f;

	for (int t = 0; t < map.n_things; t++)
	{
		if (map.things[t]->type == 1)
		{
			pos.x = map.things[t]->x * scale;
			pos.y = map.things[t]->y * scale;

			int sector = determine_sector(map.things[t]->x, map.things[t]->y);
			
			if (sector != -1)
				pos.z = (map.sectors[sector]->f_height + 40) * scale;

			if (map.things[t]->angle == 0)			// east
				view.set(1.0f, 0.0f, 0.0f);
			else if (map.things[t]->angle == 45)	// northeast
				view.set(0.7f, 0.7f, 0.0f);
			else if (map.things[t]->angle == 90)	// north
				view.set(0.0f, 1.0f, 0.0f);
			else if (map.things[t]->angle == 135)	// northwest
				view.set(-0.7f, 0.7f, 0.0f);
			else if (map.things[t]->angle == 180)	// west
				view.set(-1.0f, 0.0f, 0.0f);
			else if (map.things[t]->angle == 225)	// southwest
				view.set(-0.7f, -0.7f, 0.0f);
			else if (map.things[t]->angle == 270)	// south
				view.set(0.0f, -1.0f, 0.0f);
			else if (map.things[t]->angle == 315)	// southeast
				view.set(0.7f, -0.7f, 0.0f);
			else
				view.set(1.0f, 0.0f, 0.0f);
		}
	}
	
	camera.position.set(pos);
	camera.view.set(pos + view);
	map_changelevel(3);

	update_grid();
	update_map();
}

// close_map: Closes the open map and resets editor variables
// ------------------------------------------------------- >>
void close_map()
{
	xoff = 0;
	yoff = 0;
	gridsize = 256;
	zoom = 8;
	edit_mode = 1;
}

// add_move_items: Adds selected or hilighted item(s) to the move list
// ---------------------------------------------------------------- >>
void add_move_items()
{
	point2_t offset;
	bool	selection = false;

	// Vertices
	if (edit_mode == 0)
	{
		// Move selected vertices
		for (DWORD v = 0; v < map.n_verts; v++)
		{
			if (SELECTED(verts, v))
			{
				offset.x = map.verts[v]->x - m_x(mouse.x);
				offset.y = -map.verts[v]->y - m_y(mouse.y);

				move_list.add(offset, v);
				move_list.cursor_snap = true;

				map.verts[v]->moving = true;
				
				move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
											(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
				
				selection = true;
			}
		}

		if (selection)
			make_backup(false, false, true, false, false);
		
		// If no selection, but something is highlighted, move that
		if (!selection && hilight_item != -1)
		{
			make_backup(false, false, true, false, false);
			v = hilight_item;
			
			offset.x = map.verts[v]->x - m_x(mouse.x);
			offset.y = -map.verts[v]->y - m_y(mouse.y);
			
			move_list.add(offset, v);
			move_list.cursor_snap = false;
			
			map.verts[v]->moving = true;
		}
	}
	
	// Lines
	if (edit_mode == 1)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			// Move selected lines
			if (SELECTED(lines, l))
			{
				rect_t r = map.l_getrect(l);

				// Add vertex 1
				offset.x = r.x1() - m_x(mouse.x);
				offset.y = -r.y1() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex1);
				
				// Add vertex 2
				offset.x = r.x2() - m_x(mouse.x);
				offset.y = -r.y2() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex2);

				move_list.cursor_snap = true;
				
				map.lines[l]->moving = true;
				
				move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
											(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
				
				selection = true;
			}
		}

		if (selection)
			make_backup(false, false, true, false, false);

		// If no selection, but something is highlighted, move that
		if (!selection && hilight_item != -1)
		{
			make_backup(false, false, true, false, false);
			l = hilight_item;
			rect_t r = map.l_getrect(l);
			
			// Add vertex 1
			offset.x = r.x1() - m_x(mouse.x);
			offset.y = -r.y1() - m_y(mouse.y);
			move_list.add(offset, map.lines[l]->vertex1);
			
			// Add vertex 2
			offset.x = r.x2() - m_x(mouse.x);
			offset.y = -r.y2() - m_y(mouse.y);
			move_list.add(offset, map.lines[l]->vertex2);

			move_list.cursor_snap = true;
			
			map.lines[l]->moving = true;
			
			move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
										(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
		}
	}

	// Sectors
	if (edit_mode == 2)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			short sector1 = map.l_getsector1(l);
			short sector2 = map.l_getsector2(l);

			if (map.s_isselected(sector1))
			{
				rect_t r = map.l_getrect(l);

				// Add vertex 1
				offset.x = r.x1() - m_x(mouse.x);
				offset.y = -r.y1() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex1);
				
				// Add vertex 2
				offset.x = r.x2() - m_x(mouse.x);
				offset.y = -r.y2() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex2);

				move_list.cursor_snap = true;
				
				map.sectors[sector1]->moving = true;
				
				move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
											(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
				
				selection = true;
			}

			if (map.s_isselected(sector2))
			{
				rect_t r = map.l_getrect(l);

				// Add vertex 1
				offset.x = r.x1() - m_x(mouse.x);
				offset.y = -r.y1() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex1);
				
				// Add vertex 2
				offset.x = r.x2() - m_x(mouse.x);
				offset.y = -r.y2() - m_y(mouse.y);
				move_list.add(offset, map.lines[l]->vertex2);

				move_list.cursor_snap = true;
				
				map.sectors[sector2]->moving = true;
				
				move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
											(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
				
				selection = true;
			}
		}

		if (!selection)
		{
			for (DWORD l = 0; l < map.n_lines; l++)
			{
				short sector1 = map.l_getsector1(l);
				short sector2 = map.l_getsector2(l);

				if (map.s_ishilighted(sector1))
				{
					rect_t r = map.l_getrect(l);

					// Add vertex 1
					offset.x = r.x1() - m_x(mouse.x);
					offset.y = -r.y1() - m_y(mouse.y);
					move_list.add(offset, map.lines[l]->vertex1);
					
					// Add vertex 2
					offset.x = r.x2() - m_x(mouse.x);
					offset.y = -r.y2() - m_y(mouse.y);
					move_list.add(offset, map.lines[l]->vertex2);

					move_list.cursor_snap = true;
					
					map.sectors[sector1]->moving = true;
					
					move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
												(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
					
					selection = true;
				}

				if (map.s_ishilighted(sector2))
				{
					rect_t r = map.l_getrect(l);

					// Add vertex 1
					offset.x = r.x1() - m_x(mouse.x);
					offset.y = -r.y1() - m_y(mouse.y);
					move_list.add(offset, map.lines[l]->vertex1);
					
					// Add vertex 2
					offset.x = r.x2() - m_x(mouse.x);
					offset.y = -r.y2() - m_y(mouse.y);
					move_list.add(offset, map.lines[l]->vertex2);

					move_list.cursor_snap = true;
					
					map.sectors[sector2]->moving = true;
					
					move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
												(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
					
					selection = true;
				}
			}
		}

		if (selection)
			make_backup(false, false, true, false, false);
	}
	
	// Things
	if (edit_mode == 3)
	{
		// Move selected things
		for (DWORD t = 0; t < map.n_things; t++)
		{
			if (SELECTED(things, t))
			{
				offset.x = map.things[t]->x - m_x(mouse.x);
				offset.y = -map.things[t]->y - m_y(mouse.y);
				
				move_list.add(offset, t);
				move_list.cursor_snap = true;

				map.things[t]->moving = true;
				
				move_list.cursor_offset.set((m_x(mouse.x) - snap_to_grid(m_x(mouse.x))),
											(m_y(mouse.y) - snap_to_grid(m_y(mouse.y))));
				
				selection = true;
			}
		}

		if (selection)
			make_backup(false, false, false, false, true);
		
		// If no selection, but something is highlighted, move that
		if (!selection && hilight_item != -1)
		{
			make_backup(false, false, false, false, true);
			t = hilight_item;
			
			offset.x = map.things[t]->x - m_x(mouse.x);
			offset.y = -map.things[t]->y - m_y(mouse.y);
			
			move_list.add(offset, t);
			move_list.cursor_snap = false;

			map.things[t]->moving = true;
		}
	}
}

// check_vertex_split: Checks if a vertex is close enough to a line to split it
// ------------------------------------------------------------------------- >>
int check_vertex_split(DWORD vertex)
{
	int split_line = -1;
	short vx = map.verts[vertex]->x;
	short vy = map.verts[vertex]->y;
	
	for (DWORD l = 0; l < map.n_lines; l++)
	{
		rect_t r = map.l_getrect(l);
		
		if (distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), vx, vy) < 2
			&& map.lines[l]->vertex1 != vertex
			&& map.lines[l]->vertex2 != vertex)
			return l;
	}
	
	return -1;
}

// check_vertex_split: Checks if a point is close enough to a line to split it
// ------------------------------------------------------------------------ >>
int check_vertex_split(point2_t p)
{
	int split_line = -1;
	int vertex = map.v_getvertatpoint(p);
	
	for (DWORD l = 0; l < map.n_lines; l++)
	{
		rect_t r = map.l_getrect(l);
		
		if (distance_to_line(r.x1(), r.y1(), r.x2(), r.y2(), p.x, p.y) < 2
			&& map.lines[l]->vertex1 != vertex
			&& map.lines[l]->vertex2 != vertex)
			return l;
	}
	
	return -1;
}

// check_overlapping_lines: Checks if lines in a list are overlapping any map lines
// ----------------------------------------------------------------------------- >>
bool check_overlapping_lines(numlist_t *lines)
{
	for (DWORD a = 0; a < lines->n_numbers; a++)
	{
		int l2 = lines->numbers[a];

		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (l != l2)
			{
				if (map.lines[l]->vertex1 == map.lines[l2]->vertex1 &&
					map.lines[l]->vertex2 == map.lines[l2]->vertex2)
					return true;

				if (map.lines[l]->vertex1 == map.lines[l2]->vertex1 &&
					map.lines[l]->vertex2 == map.lines[l2]->vertex1)
					return true;
			}
		}
	}

	return false;
}

// merge_overlapping_lines: Merges any overlapping lines (only if they are in the list given)
// --------------------------------------------------------------------------------------- >>
void merge_overlapping_lines(numlist_t *lines)
{
	for (DWORD a = 0; a < lines->n_numbers; a++)
	{
		int l2 = lines->numbers[a];

		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (l != l2)
			{
				if (map.lines[l]->vertex1 == map.lines[l2]->vertex1 &&
					map.lines[l]->vertex2 == map.lines[l2]->vertex2)
				{
					int sector1 = get_side_sector(l, 1);
					int sector2 = get_side_sector(l, 2);

					if (sector1 != -1)
						map.l_setsector(l, 1, sector1);
					if (sector2 != -1)
						map.l_setsector(l, 2, sector2);

					map.delete_line(l2);
					break;
				}

				if (map.lines[l]->vertex1 == map.lines[l2]->vertex1 &&
					map.lines[l]->vertex2 == map.lines[l2]->vertex1)
				{
					int sector1 = get_side_sector(l, 1);
					int sector2 = get_side_sector(l, 2);

					if (sector1 != -1)
						map.l_setsector(l, 1, sector1);
					if (sector2 != -1)
						map.l_setsector(l, 2, sector2);

					map.delete_line(l2);
					break;
				}
			}
		}
	}

	update_map();
	map_changelevel(3);
}

// clear_move_items: Clears the move list and unflags items
// ----------------------------------------------------- >>
void clear_move_items()
{
	bool merge = false;

	move_list.clear();
	
	// Vertices
	if (edit_mode == 0)
	{
		for (DWORD v = 0; v < map.n_verts; v++)
		{
			if (!selection() && MOVING(verts, v))
			{
				int split_line = check_vertex_split(v);
				
				if (split_line != -1)
					map.l_split(split_line, v);

				if (!map.v_checkspot(map.verts[v]->x, map.verts[v]->y))
				{
					map.v_mergespot(map.verts[v]->x, map.verts[v]->y);
					merge = true;
				}
			}
			
			map.verts[v]->moving = false;
		}

		map_changed = 3;
	}
	
	// Lines
	if (edit_mode == 1)
	{
		numlist_t moving_lines;

		// Get a list of the moving lines
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (map.lines[l]->moving)
				moving_lines.add(l, true);
		}

		// Split any lines that need splitting
		for (DWORD a = 0; a < moving_lines.n_numbers; a++)
		{
			int line = moving_lines.numbers[a];

			// Vertex 1
			int split_line = check_vertex_split(map.lines[line]->vertex1);

			if (split_line != -1 && split_line != line)
				map.l_split(split_line, map.lines[line]->vertex1);

			// Vertex 2
			split_line = check_vertex_split(map.lines[line]->vertex2);

			if (split_line != -1 && split_line != line)
				map.l_split(split_line, map.lines[line]->vertex2);
		}

		// Check if we have any overlapping lines, if so merge them
		if (check_overlapping_lines(&moving_lines))
		{
			make_backup(true, true, false, false, false);
			merge_overlapping_lines(&moving_lines);
		}

		for (DWORD l = 0; l < map.n_lines; l++)
			map.lines[l]->moving = false;

		map_changed = 3;
	}
	
	// Sectors
	if (edit_mode == 2)
	{
		for (DWORD s = 0; s < map.n_sectors; s++)
			map.sectors[s]->moving = false;

		map_changed = 3;
	}
	
	// Things
	if (edit_mode == 3)
	{
		for (DWORD t = 0; t < map.n_things; t++)
			map.things[t]->moving = false;

		map_changed = 2;
	}
	
	// Remove any free vertices or zerolength lines after merging
	// vertices (in case we are left with some)
	if (merge)
	{
		remove_zerolength_lines();

		if (edit_mode != 0)
			remove_free_verts();
	}

	update_map();
}

// move_items: Moves any items in the move list
// ----------------------------------------- >>
void move_items()
{
	// Vertices
	if (edit_mode == 0)
	{
		for (DWORD v = 0; v < map.n_verts; v++)
		{
			if (MOVING(verts, v))
			{
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));
				map_changed = 2;
			}
		}
	}
	
	// Lines
	if (edit_mode == 1)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (MOVING(lines, l))
			{
				int v = map.lines[l]->vertex1;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				v = map.lines[l]->vertex2;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				map_changed = 2;
			}
		}
	}
	
	// Sectors
	if (edit_mode == 2)
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			short sector1 = map.l_getsector1(l);
			short sector2 = map.l_getsector2(l);

			if (map.s_ismoving(sector1))
			{
				int v = map.lines[l]->vertex1;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				v = map.lines[l]->vertex2;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				map_changed = 2;
			}

			if (map.s_ismoving(sector2))
			{
				int v = map.lines[l]->vertex1;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				v = map.lines[l]->vertex2;
				map.verts[v]->x = move_list.get_x(v, m_x(mouse.x));
				map.verts[v]->y = -move_list.get_y(v, m_y(mouse.y));

				map_changed = 2;
			}
		}
	}
	
	// Things
	if (edit_mode == 3)
	{
		for (DWORD t = 0; t < map.n_things; t++)
		{
			if (MOVING(things, t))
			{
				map.things[t]->x = move_list.get_x(t, m_x(mouse.x));
				map.things[t]->y = -move_list.get_y(t, m_y(mouse.y));

				map_changed = 1;
			}
		}
	}

	update_map();
}

// create_vertex: Adds a vertex at the mouse pointer and splits a line if it's close enough
// ------------------------------------------------------------------------------------- >>
void create_vertex()
{
	point2_t point(snap_to_grid(m_x(mouse.x)), snap_to_grid(-m_y(mouse.y)));

	if (map.v_checkspot(point.x, point.y))
	{
		int split_line = check_vertex_split(point);

		if (split_line != -1)
		{
			make_backup(true, true, true, false, false);
			map.add_vertex(point.x, point.y);
			map.l_split(split_line, map.n_verts - 1);
		}
		else
		{
			make_backup(false, false, true, false, false);
			map.add_vertex(point.x, point.y);
		}
	}
	else
		return;

	map_changelevel(2);
	update_map();
}

// create_lines: Creates lines between vertices in the select_order array
// (joins the last vertex to the first if close is true)
// ------------------------------------------------------------------- >>
void create_lines(bool close)
{
	if (select_order.n_numbers < 2)
		return;
	
	make_backup(true, false, false, false, false);
	numlist_t lines;
	
	for (WORD p = 0; p < select_order.n_numbers - 1; p++)
		lines.add(map.add_line(select_order.numbers[p], select_order.numbers[p + 1]), true);
	
	if (close)
		lines.add(map.add_line(select_order.numbers[select_order.n_numbers - 1], select_order.numbers[0]), true);
	
	clear_selection();
	edit_mode = 1;
	
	for (WORD l = 0; l < lines.n_numbers; l++)
		map.lines[lines.numbers[l]]->selected = true;

	lines.clear();
	
	map_changelevel(3);
	update_map();
}

// create_sector: Creates a sector out of selected lines
// -------------------------------------------------- >>
void create_sector()
{
	int sector = map.add_sector();
	int side = -1;
	make_backup(true, true, false, true, false);

	for (DWORD l = 0; l < map.n_lines; l++)
	{
		if (SELECTED(lines, l))
		{
			side = map.add_side();
			map.sides[side]->sector = sector;

			if (map.lines[l]->side1 == -1)
				map.lines[l]->side1 = side;
			else if (map.lines[l]->side2 == -1)
			{
				map.lines[l]->side2 = side;
				map.lines[l]->flags = LINE_TWOSIDED;

				memcpy(map.sectors[sector], map.sectors[map.sides[map.lines[l]->side1]->sector], sizeof(sector_t));
			}

			if (map.lines[l]->side2 == -1)
				map.lines[l]->flags = LINE_IMPASSIBLE;

			map.l_setdeftextures(l);
		}
	}

	map_changelevel(3);
	update_map();
}

// create_thing: Creates a thing at the mouse pointer, with properties taken from the last thing edited
// ------------------------------------------------------------------------------------------------- >>
void create_thing()
{
	make_backup(false, false, false, false, true);
	map.add_thing(snap_to_grid(m_x(mouse.x)), snap_to_grid(-m_y(mouse.y)), last_thing);
	update_map();
	map_changelevel(2);
}

// delete_vertex: Deletes the currently hilighted vertex or all selected vertices
// --------------------------------------------------------------------------- >>
void delete_vertex()
{
	make_backup(true, false, true, false, false);

	if (selection())
	{
		for (DWORD v = 0; v < map.n_verts; v++)
		{
			if (SELECTED(verts, v))
			{
				map.delete_vertex(v);
				v--;
			}
		}
	}
	else
	{
		if (hilight_item != -1)
			map.delete_vertex(hilight_item);
	}

	//remove_free_verts();
	map_changelevel(3);
	update_map();
}

// delete_line: Deletes the currently hilighted line or all selected lines
// -------------------------------------------------------------------- >>
void delete_line()
{
	make_backup(true, false, true, false, false);

	if (selection())
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (SELECTED(lines, l))
			{
				map.delete_line(l);
				l--;
			}
		}
	}
	else
	{
		if (hilight_item != -1)
			map.delete_line(hilight_item);
	}

	remove_free_verts();
	map_changelevel(3);
	update_map();
}

// delete_thing: Deletes the currently hilighted thing or all selected things
// ----------------------------------------------------------------------- >>
void delete_thing()
{
	make_backup(false, false, false, false, true);

	if (selection())
	{
		for (DWORD t = 0; t < map.n_things; t++)
		{
			if (SELECTED(things, t))
			{
				map.delete_thing(t);
				t--;
			}
		}
	}
	else
	{
		if (hilight_item != -1)
			map.delete_thing(hilight_item);
	}

	map_changelevel(2);
	update_map();
}

// delete_sector: Deletes the currently hilighted sector or all selected sectors
// -------------------------------------------------------------------------- >>
void delete_sector()
{
	make_backup(true, true, false, true, false);

	if (selection())
	{
		for (DWORD s = 0; s < map.n_sectors; s++)
		{
			if (SELECTED(things, s))
			{
				map.delete_sector(s);
				s--;
			}
		}
	}
	else
	{
		if (hilight_item != -1)
			map.delete_sector(hilight_item);
	}

	map_changelevel(3);
	update_map();
}

// line_flip: Flips any hilighted/selected lines
// ------------------------------------------ >>
void line_flip(bool verts, bool sides)
{
	if (selection())
	{
		for (DWORD l = 0; l < map.n_lines; l++)
		{
			if (SELECTED(lines, l))
			{
				if (verts)
					map.l_flip(l);
				if (sides)
					map.l_flipsides(l);
			}
		}
	}
	else if (hilight_item != -1)
	{
		if (verts)
			map.l_flip(hilight_item);
		if (sides)
			map.l_flipsides(hilight_item);
	}

	if (verts)
		map_changelevel(3);
	else
		map_changelevel(2);

	update_map();
}

// sector_changeheight: Changes floor/ceiling height of hilighted or selected sector(s)
// --------------------------------------------------------------------------------- >>
void sector_changeheight(bool floor, int amount)
{
	if (!selection())
	{
		if (hilight_item != -1)
			map.s_changeheight(hilight_item, floor, amount);
	}
	else
	{
		for (DWORD s = 0; s < map.n_sectors; s++)
		{
			if (SELECTED(sectors, s))
				map.s_changeheight(s, floor, amount);
		}
	}

	map_changelevel(2);
}

void merge_verts()
{
	for (DWORD v = 0; v < map.n_verts; v++)
		map.v_mergespot(map.verts[v]->x, map.verts[v]->y);

	map_changelevel(3);
	//remove_free_verts();
}

bool lines_clockwise(numlist_t *lines)
{
	for (DWORD a = 0; a < lines->n_numbers; a++)
	{
		int l = lines->numbers[a];

		rect_t linerect = map.l_getrect(l);
		point2_t mid = midpoint(linerect.tl, linerect.br);
		point2_t vec(linerect.x2() - linerect.x1(), linerect.y2() - linerect.y1());
		int x = vec.y;
		int y = vec.x;
		x = -x;
		point2_t side(mid.x + x, mid.y + y);
		
		bool intersect = false;

		for (DWORD b = 0; b < lines->n_numbers; b++)
		{
			if (b != a)
			{
				rect_t line = map.l_getrect(lines->numbers[b]);

				int x1 = line.x1();
				int x2 = line.x2();
				int y1 = line.y1();
				int y2 = line.y2();

				point2_t r1 = mid;
				point2_t r2 = side;

				float u_ray = (float(x2 - x1) * float(r1.y - y1) - float(y2 - y1) * float(r1.x - x1)) /
							(float(y2 - y1) * float(r2.x - r1.x) - float(x2 - x1) * float(r2.y - r1.y));
				
				float u_line = (float(r2.x - r1.x) * float(r1.y - y1) - float(r2.y - r1.y) * float(r1.x - x1)) /
							(float(y2 - y1) * float(r2.x - r1.x) - float(x2 - x1) * float(r2.y - r1.y));
				
				if((u_ray >= 0)/* && (u_ray <= 1024)*/ && (u_line >= 0) && (u_line <= 1))
					intersect = true;
			}
		}

		if (!intersect)
			return true;
	}

	return false;
}

bool check_split(int split_sector)
{
	// Cycle through lines starting with the end vertex, and if we get back to the end
	// vertex before the start vertex, it means we don't have a proper sector split
	numlist_t checked_lines, a_lines, prev_verts;
	int vertex = map.v_getvertatpoint(ldraw_points.get_last());
	int end_vert = map.v_getvertatpoint(ldraw_points.points[0]);
	int start = vertex;
	prev_verts.add(start, false);
	bool done = false;

	while (!done)
	{
		// Get any lines attached to the current vertex
		a_lines.clear();
		map.v_getattachedlines(vertex, &a_lines);

		for (DWORD l = 0; l < a_lines.n_numbers; l++)
		{
			if (!checked_lines.exists(a_lines.numbers[l]) &&
				!ldraw_lines.exists(a_lines.numbers[l]))
			{
				// Ignore lines with both sides referencing our sector
				if (map.l_getsector1(a_lines.numbers[l]) == split_sector
					&& map.l_getsector2(a_lines.numbers[l]) == split_sector)
				{
					checked_lines.add(a_lines.numbers[l], true);
					l++;
				}

				if (map.lines[a_lines.numbers[l]]->vertex1 == vertex &&
					map.l_getsector1(a_lines.numbers[l]) == split_sector)
				{
					vertex = map.lines[a_lines.numbers[l]]->vertex2;
					checked_lines.add(a_lines.numbers[l], true);
					prev_verts.add(vertex, false);

					if (vertex == end_vert)
						return true;

					// If we're back at the start vertex, go to the previous vertex
					if (vertex == start)
					{
						vertex = prev_verts.get_last();
						prev_verts.remove_last();

						if (prev_verts.n_numbers == 0)
							return false;
					}

					break;
				}

				if (map.lines[a_lines.numbers[l]]->vertex2 == vertex &&
					map.l_getsector2(a_lines.numbers[l]) == split_sector)
				{
					vertex = map.lines[a_lines.numbers[l]]->vertex1;
					checked_lines.add(a_lines.numbers[l], true);
					prev_verts.add(vertex, false);

					if (vertex == end_vert)
						return true;

					if (vertex == start)
					{
						vertex = prev_verts.get_last();
						prev_verts.remove_last();

						if (prev_verts.n_numbers == 0)
							return false;
					}

					break;
				}
			}

			// If we're at the end of the attached lines without
			// coming across any not already checked or part of the sector,
			// either go back a vertex (if we're not at the start), or exit
			// the loop (and return false)
			if (l == a_lines.n_numbers - 1)
			{
				if (vertex == start)
					done = true;
				else
				{
					vertex = prev_verts.get_last();
					prev_verts.remove_last();

					if (prev_verts.n_numbers == 0)
						return false;
				}
			}
		}
	}

	return false;
}

// end_linedraw: Finishes a line drawing session, builds lines/vertices/sectors
// ------------------------------------------------------------------------- >>
void end_linedraw()
{
	make_backup(true, true, true, true, false);

	if (line_draw && ldraw_points.n_points > 2) // Line drawing finished without error/cancellation
	{
		linelist_t		new_lines;

		bool split = false;

		// If the last 2 points are the same, we have a sector split
		if (ldraw_points.get_last().equals(ldraw_points.points[ldraw_points.n_points - 2]))
			split = true;

		// *** SPLIT SECTOR ***
		// I'm sure theres a much nicer way to do this, but the method I use actually works,
		// So it'll do for now :P
		if (split)
		{
			numlist_t a_lines; // List of lines attached to vertex being examined

			// Determine sector to split, if any.
			// Do so by finding the nearest sector to the midpoint of the last line drawn
			point2_t mid = midpoint(ldraw_points.points[0], ldraw_points.points[1]);
			int split_sector = determine_sector(mid.x, mid.y);

			// Add a vertex at each point, also add vertex numbers to the created vertices list
			// Also, split any lines if need be, and ignore duplicate points
			for (int p = 0; p < ldraw_points.n_points; p++)
			{
				if (!ldraw_points.points[p].equals(ldraw_points.points[p - 1]))
					ldraw_verts.add(map.add_vertex(ldraw_points.points[p].x, ldraw_points.points[p].y), true);

				int split_line = check_vertex_split(map.n_verts - 1);

				if (split_line != -1)
					map.l_split(split_line, map.n_verts - 1);
			}

			// Add lines from the vertices, also adding line numbers to the created lines list
			for (int v = 1; v < ldraw_verts.n_numbers; v++)
				ldraw_lines.add(map.add_line(ldraw_verts.numbers[v - 1], ldraw_verts.numbers[v]), true);

			// Merge vertices (in the entire map, for the moment)
			merge_verts();
			remove_free_verts();

			// Rebuild vertices list
			ldraw_verts.clear();
			for (DWORD l = 0; l < new_lines.n_lines; l++)
				ldraw_verts.add(map.lines[ldraw_lines.numbers[l]]->vertex1, true);

			ldraw_verts.add(map.lines[ldraw_lines.get_last()]->vertex2, true);

			// If the split line(s) has/have a loose end
			if (!map.v_isattached_sector(map.v_getvertatpoint(ldraw_points.points[0]))
				|| !map.v_isattached_sector(map.v_getvertatpoint(ldraw_points.get_last())))
			{
				// Run through newly created lines, set both their sides to the 'split' sector
				for (DWORD l = 0; l < ldraw_lines.n_numbers; l++)
				{
					map.l_setsector(ldraw_lines.numbers[l], 1, split_sector);
					map.l_setsector(ldraw_lines.numbers[l], 2, split_sector);
					map.lines[ldraw_lines.numbers[l]]->set_flag(LINE_TWOSIDED);
				}
			}
			else if (!check_split(split_sector))
			{
				//printf("check_split failed\n");

				// Run through newly created lines, set both their sides to the 'split' sector
				for (DWORD l = 0; l < ldraw_lines.n_numbers; l++)
				{
					map.l_setsector(ldraw_lines.numbers[l], 1, split_sector);
					map.l_setsector(ldraw_lines.numbers[l], 2, split_sector);
					map.lines[ldraw_lines.numbers[l]]->set_flag(LINE_TWOSIDED);
				}
			}
			else
			{
				// Create a new sector
				int new_sector = map.add_sector();

				// Copy split sector attributes to new sector
				if (split_sector != -1)
					memcpy(map.sectors[new_sector], map.sectors[split_sector], sizeof(sector_t));

				// Run through newly created lines, set their first side to the new
				// sector, and their second side to the split sector
				numlist_t changed_lines;

				for (DWORD l = 0; l < ldraw_lines.n_numbers; l++)
				{
					map.l_setsector(ldraw_lines.numbers[l], 1, new_sector);
					map.l_setsector(ldraw_lines.numbers[l], 2, split_sector);
					map.lines[ldraw_lines.numbers[l]]->set_flag(LINE_TWOSIDED);
					changed_lines.add(ldraw_lines.numbers[l], true);
				}

				// Starting at the last new vertex, cycle through lines that are part of
				// the sector to be split, and set whatever side's sector reference
				// to the new sector. End once we arrive at the first created vertex. Phew.
				bool done = false;
				bool not_split = false;
				int vertex = map.v_getvertatpoint(ldraw_points.get_last());
				int end_vert = map.v_getvertatpoint(ldraw_points.points[0]);
				int start_vert = vertex;

				while (!done)
				{
					// Get any lines attached to the current vertex
					map.v_getattachedlines(vertex, &a_lines);

					for (DWORD l = 0; l < a_lines.n_numbers; l++)
					{
						if (!ldraw_lines.exists(a_lines.numbers[l]) &&
							!changed_lines.exists(a_lines.numbers[l]))
						{
							// If the line has both sides pointing to the sector to be split, skip it
							// We'll deal with these lines later
							if (map.l_getsector1(a_lines.numbers[l]) == split_sector
								&& map.l_getsector2(a_lines.numbers[l]) == split_sector)
								l++;

							if (map.lines[a_lines.numbers[l]]->vertex1 == vertex &&
								map.l_getsector1(a_lines.numbers[l]) == split_sector)
							{
								map.l_setsector(a_lines.numbers[l], 1, new_sector);
								vertex = map.lines[a_lines.numbers[l]]->vertex2;
								changed_lines.add(a_lines.numbers[l], true);

								if (vertex == end_vert)
									done = true;

								break;
							}

							if (map.lines[a_lines.numbers[l]]->vertex2 == vertex &&
								map.l_getsector2(a_lines.numbers[l]) == split_sector)
							{
								map.l_setsector(a_lines.numbers[l], 2, new_sector);
								vertex = map.lines[a_lines.numbers[l]]->vertex1;
								changed_lines.add(a_lines.numbers[l], true);

								if (vertex == end_vert)
									done = true;

								break;
							}
						}

						// To prevent an infinite loop
						if (l == a_lines.n_numbers - 1)
							done = true;
					}
				}

				// Reverse cycle (just to add some changed_lines for later)
				done = false;
				vertex = map.v_getvertatpoint(ldraw_points.get_last());
				start_vert = vertex;

				while (!done)
				{
					// Get any lines attached to the current vertex
					map.v_getattachedlines(vertex, &a_lines);

					for (DWORD l = 0; l < a_lines.n_numbers; l++)
					{
						if (!ldraw_lines.exists(a_lines.numbers[l]) &&
							!changed_lines.exists(a_lines.numbers[l]))
						{
							// If the line has both sides pointing to the sector to be split, skip it
							// We'll deal with these lines later
							if (map.l_getsector1(a_lines.numbers[l]) == split_sector
								&& map.l_getsector2(a_lines.numbers[l]) == split_sector)
								l++;

							if (map.lines[a_lines.numbers[l]]->vertex2 == vertex &&
								map.l_getsector1(a_lines.numbers[l]) == split_sector
								&& !changed_lines.exists(a_lines.numbers[l]))
							{
								vertex = map.lines[a_lines.numbers[l]]->vertex1;
								changed_lines.add(a_lines.numbers[l], true);

								if (vertex == end_vert)
									done = true;

								break;
							}

							if (map.lines[a_lines.numbers[l]]->vertex1 == vertex &&
								map.l_getsector2(a_lines.numbers[l]) == split_sector
								&& !changed_lines.exists(a_lines.numbers[l]))
							{
								vertex = map.lines[a_lines.numbers[l]]->vertex2;
								changed_lines.add(a_lines.numbers[l], true);

								if (vertex == end_vert)
									done = true;

								break;
							}
						}

						// To prevent an infinite loop
						if (l == a_lines.n_numbers - 1)
							done = true;
					}
				}

				if (not_split)
				{
					for (DWORD l = 0; l < ldraw_lines.n_numbers; l++)
					{
						map.l_setsector(ldraw_lines.numbers[l], 1, split_sector);
						map.l_setsector(ldraw_lines.numbers[l], 2, split_sector);
					}
				}
				else
				{
					// Now we go through all the lines that are part of the sector to be split
					// and determine which side of the split they are on
					for (l = 0; l < map.n_lines; l++)
					{
						if ((map.l_getsector1(l) == split_sector || map.l_getsector2(l) == split_sector)
							&& !changed_lines.exists(l) && !ldraw_lines.exists(l))
						{
							int closest_line = -1;
							int min_dist = -1;

							rect_t line_rect = map.l_getrect(l);
							float mid_x = (float)midpoint(line_rect.tl, line_rect.br).x;
							float mid_y = (float)midpoint(line_rect.tl, line_rect.br).y;

							if (map.l_getsector1(l) == split_sector)
							{
								if (line_rect.x1() > line_rect.x2())
									mid_y += 0.5f;
								if (line_rect.x2() > line_rect.x1())
									mid_y -= 0.5f;
								if (line_rect.y1() > line_rect.y2())
									mid_x -= 0.5f;
								if (line_rect.y2() > line_rect.y1())
									mid_x += 0.5f;
							}

							if (map.l_getsector2(l) == split_sector)
							{
								if (line_rect.x1() > line_rect.x2())
									mid_y -= 0.5f;
								if (line_rect.x2() > line_rect.x1())
									mid_y += 0.5f;
								if (line_rect.y1() > line_rect.y2())
									mid_x += 0.5f;
								if (line_rect.y2() > line_rect.y1())
									mid_x -= 0.5f;
							}

							point2_t mid(midpoint(map.l_getrect(l).tl, map.l_getrect(l).br));

							for (DWORD a = 0; a < changed_lines.n_numbers; a++)
							{
								rect_t line2(map.l_getrect(changed_lines.numbers[a]).tl, map.l_getrect(changed_lines.numbers[a]).br);
								int dist = distance_to_line(line2.x1(), line2.y1(), line2.x2(), line2.y2(), mid.x, mid.y);

								if (min_dist == -1 || dist < min_dist)
								{
									min_dist = dist;
									closest_line = changed_lines.numbers[a];
								}
							}

							// Now figure out what sector the line should be in
							int sector = -1;

							if (map.l_getsector1(l) == split_sector)
							{
								if (determine_line_side(closest_line, mid_x, mid_y))
								{
									if (map.l_getsector1(closest_line) == new_sector)
										map.l_setsector(l, 1, new_sector);
								}
								else
								{
									if (map.l_getsector2(closest_line) == new_sector)
										map.l_setsector(l, 1, new_sector);
								}
							}

							if (map.l_getsector2(l) == split_sector)
							{
								if (determine_line_side(closest_line, mid_x, mid_y))
								{
									if (map.l_getsector1(closest_line) == new_sector)
										map.l_setsector(l, 2, new_sector);
								}
								else
								{
									if (map.l_getsector2(closest_line) == new_sector)
										map.l_setsector(l, 2, new_sector);
								}
							}
						}
					}
				}
			}
		}
		else // *** CLOSED SECTOR ADDED ***
		{
			bool new_sector_used = false;
			bool new_sector_copied = false;
			int new_sector = map.add_sector();
			numlist_t new_lines;
			//numlist_t changed_sides;
			numlist_t check_sectors;
			numlist_t changed_lines;

			// TODO: Check if sector splits are adequate

			// Go through all new points
			for (DWORD p = 0; p < ldraw_points.n_points - 1; p++)
			{
				bool done = false;
				point2_t p1 = ldraw_points.points[p];
				point2_t p2 = ldraw_points.points[p + 1];
				point2_t p3;
				int v1 = -1;
				int v2 = -1;

				if (p == 0)
					p3 = ldraw_points.points[p + 2];
				else
					p3 = ldraw_points.points[p - 1];

				// If we have a point on a line that isn't a current vertex,
				// add the vertex and split the line
				if (map.v_checkspot(p1.x, p1.y) && check_vertex_split(p1) != -1)
				{
					v1 = map.add_vertex(p1.x, p1.y);
					map.l_split(check_vertex_split(p1), v1);
				}
				else
					v1 = map.v_getvertatpoint(p1);

				if (map.v_checkspot(p2.x, p2.y) && check_vertex_split(p2) != -1)
				{
					v2 = map.add_vertex(p2.x, p2.y);
					map.l_split(check_vertex_split(p2), v2);
				}
				else
					v2 = map.v_getvertatpoint(p2);

				// If the two points we're looking at are both current vertices
				if (v1 != -1 && v2 != -1)
				{
					for (DWORD l = 0; l < map.n_lines; l++)
					{
						// If the points are the front side of the line
						if (map.lines[l]->vertex1 == v1 && map.lines[l]->vertex2 == v2)
						{
							// Set first side to the new sector
							if (map.l_getsector1(l) != -1 && !new_sector_copied)
							{
								memcpy(map.sectors[new_sector], map.sectors[map.sides[map.lines[l]->side1]->sector], sizeof(sector_t));
								new_sector_copied = true;
							}
							
							check_sectors.add(map.l_getsector1(l), true);
							map.l_setsector(l, 1, new_sector);

							changed_lines.add(l, true);
							done = true;
							new_sector_used = true;
							break;
						}

						// If the points are the back side of the line
						if (map.lines[l]->vertex2 == v1 && map.lines[l]->vertex1 == v2)
						{
							// Set second side to the new sector
							if (map.l_getsector1(l) != -1 && !new_sector_copied)
							{
								memcpy(map.sectors[new_sector], map.sectors[map.sides[map.lines[l]->side1]->sector], sizeof(sector_t));
								new_sector_copied = true;
							}

							check_sectors.add(map.l_getsector1(l), true);
							map.l_setsector(l, 2, new_sector);

							changed_lines.add(l, true);
							done = true;
							new_sector_used = true;
							break;
						}
					}
				}

				if (!done)
				{
					// If not, create the needed vertices and a new line
					if (v1 == -1)
						v1 = map.add_vertex(p1.x, p1.y);

					if (v2 == -1)
						v2 = map.add_vertex(p2.x, p2.y);

					int new_line = map.add_line(v1, v2);
					new_lines.add(new_line, true);
				}
			}

			// If the lines are clockwise, set their first side to the new sector
			// otherwise set their second sides to the new sector. Also set the opposite
			// side to whatever sector it's in
			numlist_t all_new_lines;
			all_new_lines.add_numlist(&new_lines, true);
			all_new_lines.add_numlist(&changed_lines, true);

			if (lines_clockwise(&all_new_lines))
			{
				for (DWORD a = 0; a < new_lines.n_numbers; a++)
				{
					int l = new_lines.numbers[a];
					int sector2 = get_side_sector(l, 2);
					check_sectors.add(sector2, true);

					if (!new_sector_copied && sector2 != -1)
					{
						memcpy(map.sectors[new_sector], map.sectors[sector2], sizeof(sector_t));
						new_sector_copied = true;
					}

					map.l_setsector(l, 1, new_sector);
					map.l_setsector(l, 2, sector2);
					map.l_setdeftextures(l);
					new_sector_used = true;
				}
			}
			else
			{
				// If we have no changed lines, it means we have a column
				if (changed_lines.n_numbers == 0)
				{
					//printf("column\n");

					for (DWORD a = 0; a < new_lines.n_numbers; a++)
					{
						int l = new_lines.numbers[a];
						int sector1 = get_side_sector(l, 1);
						check_sectors.add(sector1, true);

						map.l_setsector(l, 1, sector1);
						map.l_setdeftextures(l);
						new_sector_used = false;
					}
				}
				else
				{
					for (DWORD a = 0; a < new_lines.n_numbers; a++)
					{
						int l = new_lines.numbers[a];
						int sector2 = get_side_sector(l, 2);
						check_sectors.add(sector2, true);

						if (!new_sector_copied && sector2 != -1)
						{
							memcpy(map.sectors[new_sector], map.sectors[sector2], sizeof(sector_t));
							new_sector_copied = true;
						}

						map.l_setsector(l, 1, new_sector);
						map.l_setsector(l, 2, sector2);
						map.l_setdeftextures(l);
						new_sector_used = true;
					}
				}
			}

			// Now we process any lines that might need sector references changed
			// in a similar fashion to above (determine what side is in what sector)
			if (check_sectors.exists(-1))
				check_sectors.remove(-1);

			for (DWORD s = 0; s < check_sectors.n_numbers; s++)
			{
				int sector = check_sectors.numbers[s];

				for (DWORD l = 0; l < map.n_lines; l++)
				{
					if (!all_new_lines.exists(l))
					{
						if (map.l_getsector1(l) == sector)
						{
							int d_sector = get_side_sector(l, 1);

							if (d_sector != sector && d_sector != -1)
								map.l_setsector(l, 1, d_sector);
						}

						if (map.l_getsector2(l) == sector)
						{
							int d_sector = get_side_sector(l, 2);

							if (d_sector != sector && d_sector != -1)
								map.l_setsector(l, 2, d_sector);
						}
					}
				}
			}

			if (!new_sector_used)
				map.delete_sector(new_sector);
		}

		// Finish up
		update_map();
		ldraw_verts.clear();
		ldraw_lines.clear();
		map_changelevel(3);
	}

	line_draw = false;
	ldraw_points.clear();
}

// line_drawpoint: Adds the current mouse position to the line draw points list
// ------------------------------------------------------------------------- >>
void line_drawpoint()
{
	if (!ldraw_points.add(point2_t(snap_to_grid(m_x(mouse.x)), snap_to_grid(-m_y(mouse.y))), false))
		end_linedraw();
}

// line_undrawpoint: Removes the last entry in the line draw points list
// ------------------------------------------------------------------ >>
void line_undrawpoint()
{
	ldraw_points.remove(ldraw_points.get_last());

	if (ldraw_points.n_points == 0)
	{
		ldraw_points.clear();
		line_draw = false;
	}
}

// map_changelevel: Sets the map_changed variable, ignores the change if not applicable
// --------------------------------------------------------------------------------- >>
void map_changelevel(int level)
{
	if (level > map_changed)
		map_changed = level;
}